import logging
import odps.errors
from odps import ODPS

from migration.connector.source.enum import Column

from migration.connector.source.base import Source
from migration.base.exceptions import SourceExecutionError

logger = logging.getLogger(__name__)


class OdpsSource(Source):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('Odps', config)
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_config_path = storage_conf_path
        self.odps = ODPS(self.connection_params['id'], self.connection_params['key'],
                         self.connection_params['project'], self.connection_params['endpoint'])
        

    def get_connection_params(self):
        assert self.config['id']
        assert self.config['key']
        assert self.config['project']
        assert self.config['endpoint']

        return {
            'id': self.config['id'],
            'key': self.config['key'],
            'project': self.config['project'],
            'endpoint': self.config['endpoint'],
        }

    def connect(self):
        try:
            self.odps.execute_sql("select 1")
            logger.info(f"Connect to Odps {self.connection_params['endpoint']} successfully")
        except odps.errors.ODPSError as e:
            logger.error(f"Connect to Odps {self.connection_params['endpoint']} failed, error: {e}")
            raise f"Connect to Odps {self.connection_params['endpoint']} failed, error: {e}"

    def get_table_columns(self, database_name, table_name) -> list[Column]:
        result = self.odps.get_table(table_name)
        table_columns = []
        for row in result.columns:
            table_columns.append(Column(name=row.name, type=row.type.name.upper(),
                                        is_null=row.type.nullable,
                                        default_value=None))
        return table_columns

    def execute_sql(self, sql, bind_params=None):
        try:
            result = []
            with self.odps.execute_sql(sql).open_reader() as reader:
                for record in reader:
                    result.append(record.values)
            return result
        except odps.errors.ODPSError as e:
            logger.error(f"Odps connector execute sql {sql} failed, error: {e}")
            raise f"Odps connector execute sql {sql} failed, error: {e}"
        except SourceExecutionError as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"
        except Exception as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"

    def type_mapping(self):
        return {}

    def close(self):
        logger.info(f"Close connection to odps {self.connection_params['endpoint']} successfully")

    def int_type_string(self):
        return 'INT'

    def quote_character(self):
        return "`"



import logging

from odps import ODPS
import odps.errors

from migration.connector.destination.base import Destination
from migration.base.exceptions import DestinationExecutionError, GrammarRestrictionsError

logger = logging.getLogger(__name__)

MAX_GET_STATUS_TIMES = 100


class OdpsDestination(Destination):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('Odps', config)
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_config_path = storage_conf_path
        self.odps = ODPS(self.connection_params['id'], self.connection_params['key'],
                         self.connection_params['project'], self.connection_params['endpoint'])

    def get_connection_params(self):
        assert self.config['id']
        assert self.config['key']
        assert self.config['project']
        assert self.config['endpoint']

        return {
            'id': self.config['id'],
            'key': self.config['key'],
            'project': self.config['project'],
            'endpoint': self.config['endpoint'],
        }

    def connect(self):
        try:
            self.odps.execute_sql("select 1")
            logger.info(f"Connect to Odps {self.connection_params['endpoint']} successfully")
        except odps.errors.ODPSError as e:
            logger.error(f"Connect to Odps {self.connection_params['endpoint']} failed, error: {e}")
            raise f"Connect to Odps {self.connection_params['endpoint']} failed, error: {e}"

    def create_table(self, table_name, ddl: str):
        try:
            self.odps.execute_sql(ddl)
            logger.info(f"Odps Create table {table_name} successfully")
        except odps.errors.ODPSError as e:
            logger.error(f"Odps Create table {table_name} failed, error: {e}")
            raise f"Odps Create table {table_name} failed, error: {e}"

    def execute_sql(self, sql, bind_params=None):
        try:
            result = []
            with self.odps.execute_sql(sql).open_reader() as reader:
                for record in reader:
                    result.append(record.values)
            return result
        except odps.errors.ODPSError as e:
            logger.error(f"Odps connector execute sql {sql} failed, error: {e}")
            raise f"Odps connector execute sql {sql} failed, error: {e}"
        except DestinationExecutionError as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"
        except Exception as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"

    def close(self):
        logger.info(f"Odps connector close successfully")


"""
migration migration tool
usage:
    cz-migration test -p <profile_path> [test connection for source and destination]
    cz-migration meta -p <profile_path>  --out_path <meta_out_file_path>[generate meta from source]
    cz-migration schema -p <profile_path> [--table_list_file <external_table_list_file>][migrate schema from source to destination]
    cz-migration data -p <profile_path> --meta_conf <meta_conf> --storage_conf <storage_conf> [--table_list_file <external_table_list_file>][migrate data from source to destination]
    cz-migration validate -p <profile_path> [--table_list_file <external_table_list_file>][validate data from source and destination]
    cz-migration run -p <profile_path> --meta_conf <meta_conf> --storage_conf <storage_conf> [--table_list_file <external_table_list_file>][migrate schema and data from source to destination, including validation]
    cz-migration status -p <profile_path> [check migration status]
    cz-migration resume -p <profile_path> [resume migration]
    cz-migration report -p <profile_path> [generate migration report]
"""
import logging
import os.path
import time

import docopt
import yaml
from migration.connector.destination.base import Destination
from migration.connector.destination.clickzetta.destination import ClickZettaDestination
from migration.connector.destination.doris.destination import DorisDestination
from migration.connector.source import Source
from migration.connector.source.clickzetta.source import ClickzettaSource
from migration.connector.source.doris.source import DorisSource
from migration.scheduler.data_transformer.transformer import DataTransformer
from migration.scheduler.data_validation.validation import Validation
from migration.scheduler.schema_transformer.transformer import SchemaTransformer
from migration.scheduler.unify_transformer.transformer import UnifyTransformer

logger = logging.getLogger(__name__)


def test(source: Source, destination: Destination):
    if source.migration_test_connection() and destination.migration_test_connection():
        print("Test connection successfully!")
    else:
        print("Test connection failed!")


def meta(source: Source, out_path: str):
    logger.info(f"Generating meta from source {source} to {out_path}")
    print(f"Generating meta from source {source} to {out_path}")
    start_time = time.time()
    dbs = source.get_database_names()
    out_path_dir = os.path.dirname(out_path)
    if not os.path.exists(out_path_dir):
        os.makedirs(out_path_dir)
    with open(out_path, 'w') as f:
        for db in dbs:
            if db.startswith("__") or db.startswith("information_schema"):
                continue
            tables = source.get_table_names(db)
            for table in tables:
                f.write(f"{db}.{table}\n")

    logger.info(f"Generating meta from source {source} to {out_path} finished, time cost: {time.time() - start_time}")


def schema(source: Source, destination: Destination, db_list: list, config_table_list: list, external_table_list: list,
           project_name: str, scheduler_concurrency: int, quit_if_fail: bool, thread_concurrency: int):
    logger.info(f"Transforming schema from source {source} to destination {destination}")
    print(f"Transforming schema from source {source} to destination {destination}")
    start_time = time.time()
    SchemaTransformer(source, destination, project_name, db_list, config_table_list, external_table_list,
                      scheduler_concurrency,
                      quit_if_fail, thread_concurrency).transform()
    logger.info(
        f"Transforming schema from source {source} to destination {destination} finished, time cost: {time.time() - start_time}")


def data(source: Source, destination: Destination, db_list: list, config_table_list: list, external_table_list: list,
         project_name: str, scheduler_concurrency: int, quit_if_fail: bool, thread_concurrency: int,
         transform_partitions: dict, dest_table_list: list):
    logger.info(f"Transforming data from source {source} to destination {destination}")
    print(f"Transforming data from source {source} to destination {destination}")
    start_time = time.time()
    DataTransformer(source, destination, project_name, db_list, config_table_list, external_table_list,
                    scheduler_concurrency,
                    quit_if_fail, thread_concurrency, transform_partitions, dest_table_list).transform()
    logger.info(
        f"Transforming data from source {source} to destination {destination} finished, time cost: {time.time() - start_time}")


def validate(source: Source, destination: Destination, db_list: list, config_table_list: list,
             external_table_list: list, project_name: str, scheduler_concurrency: int, quit_if_fail: bool,
             thread_concurrency: int):
    logger.info(f"Validating data from source {source} and destination {destination}")
    print(f"Validating data from source {source} and destination {destination}")
    start_time = time.time()
    Validation(source, destination, project_name, db_list, config_table_list, external_table_list,
               scheduler_concurrency,
               quit_if_fail, thread_concurrency).transform()
    logger.info(
        f"Validating data from source {source} and destination {destination} finished, time cost: {time.time() - start_time}")


def run(source: Source, destination: Destination, db_list: list, config_table_list: list, external_table_list: list,
        project_name: str, scheduler_concurrency: int, quit_if_fail: bool, thread_concurrency: int,
        transform_partitions: dict):
    logger.info(f"Running migration from source {source} to destination {destination}")
    print(f"Running migration from source {source} to destination {destination}")
    start_time = time.time()
    UnifyTransformer(source, destination, project_name, db_list, config_table_list, external_table_list,
                     scheduler_concurrency,
                     quit_if_fail, thread_concurrency, transform_partitions).transform()
    logger.info(
        f"Running migration from source {source} to destination {destination} finished, time cost: {time.time() - start_time}")


def status(destination: Destination, project_name: str):
    import migration.util.migration_tasks_status as mts
    mts.get_last_migration_status(destination, project_name)


def resume(source: Source, destination: Destination):
    pass


def report(destination: Destination, project_name: str):
    import migration.util.migration_tasks_status as mts
    mts.get_last_migration_report(destination, project_name)


def main():
    logging.basicConfig(level=logging.INFO)
    args = docopt.docopt(__doc__)
    profile = None
    source = None
    destination = None
    concurrency = 1
    thread_concurrency = 1
    quit_if_fail = False
    meta_conf_path = None
    storage_conf_path = None
    if not args['-p']:
        print("Please specify profile path!")
        logger.error("Please specify profile path!")
    profile_path = args['<profile_path>']
    print(f"Using profile {profile_path}")
    with open(profile_path, 'r') as f:
        profile = yaml.safe_load(f)['config']
    project_name = profile['name']
    print(f"Using project name  {project_name}")
    project_version = profile['version']
    print(f"Using project version  {project_version}")
    project_desc = profile['description']
    print(f"Using project description  {project_desc}")
    source_config = profile['source']
    if args['--meta_conf']:
        meta_conf_path = args['<meta_conf>']
        print(f"Using meta conf path  {meta_conf_path}")
    if args['--storage_conf']:
        storage_conf_path = args['<storage_conf>']
        print(f"Using storage conf path  {storage_conf_path}")
    db_list = []
    config_table_list = []
    dest_table_list = None
    external_table_list = []
    transform_partitions = {}
    if source_config['type'].strip() == 'doris':
        config = {'fe_servers': source_config['fe_servers'], 'user': source_config['username'],
                  'password': source_config['password']}
        logger.info(f"migration source {config}")
        if 'transform_partitions' in source_config and source_config['transform_partitions']:
            tables = source_config['transform_partitions']['tables']
            values = source_config['transform_partitions']['values']
            if not len(values) == len(tables):
                raise Exception("transform_partitions columns, values and tables should have same length")
            for i in range(len(tables)):
                transform_partitions[tables[i]] = [values[i]]
        print(f"migration source {config}")
        source = DorisSource(config, meta_conf_path, storage_conf_path)
        if args['--table_list_file']:
            with open(args['<external_table_list_file>'], 'r') as f:
                lines = f.readlines()
                for line in lines:
                    external_table_list.append(line.strip())
        else:
            if 'migration_dbs' not in source_config:
                if 'migration_tables' not in source_config:
                    raise Exception("migration_dbs and migration_tables are both not provided in profile.yml")
                else:
                    config_table_list = source_config['migration_tables']
                    if 'dest_tables' in source_config:
                        dest_table_list = source_config['dest_tables']
                        if len(config_table_list) != len(dest_table_list):
                            raise Exception("migration_tables and dest_tables should have same length")
            else:
                db_list = source_config['migration_dbs']
    elif source_config['type'].strip() == 'clickzetta':
        config = {'service': source_config['service'], 'username': source_config['username'],
                  'workspace': source_config['workspace'], 'password': source_config['password'],
                  'instance': source_config['instance'], 'vcluster': source_config['vcluster'],
                  'instanceId': source_config['instanceId']}
        logger.info(f"migration source {config}")
        if 'transform_partitions' in source_config and source_config['transform_partitions']:
            tables = source_config['transform_partitions']['tables']
            columns = source_config['transform_partitions']['columns']
            values = source_config['transform_partitions']['values']
            if not len(columns) == len(values) == len(tables):
                raise Exception("transform_partitions columns, values and tables should have same length")
            for i in range(len(columns)):
                transform_partitions[tables[i]] = [columns[i], values[i]]
        print(f"migration source {config}")
        source = ClickzettaSource(config, meta_conf_path, storage_conf_path)
        if args['--table_list_file']:
            with open(args['<external_table_list_file>'], 'r') as f:
                lines = f.readlines()
                for line in lines:
                    external_table_list.append(line.strip())
        else:
            if 'migration_dbs' not in source_config:
                if 'migration_tables' not in source_config:
                    raise Exception("migration_dbs and migration_tables are both not provided in profile.yml")
                else:
                    config_table_list = source_config['migration_tables']
                    if 'dest_tables' in source_config:
                        dest_table_list = source_config['dest_tables']
                        if len(config_table_list) != len(dest_table_list):
                            raise Exception("migration_tables and dest_tables should have same length")
            else:
                db_list = source_config['migration_dbs']
    else:
        raise Exception(f"Unsupported source type: {source_config['type']}")
    destination_config = profile['destination']
    if destination_config['type'].strip() == 'clickzetta':
        config = {'service': destination_config['service'], 'username': destination_config['username'],
                  'workspace': destination_config['workspace'], 'password': destination_config['password'],
                  'instance': destination_config['instance'], 'vcluster': destination_config['vcluster'],
                  'instanceId': destination_config['instanceId']}
        logger.info(f"migration destination {config}")
        print(f"migration destination {config}")
        destination = ClickZettaDestination(config, meta_conf_path, storage_conf_path)
    elif destination_config['type'].strip() == 'doris':
        config = {'fe_servers': destination_config['fe_servers'], 'user': destination_config['username'],
                  'password': destination_config['password']}
        logger.info(f"migration destination {config}")
        print(f"migration destination {config}")
        destination = DorisDestination(config, meta_conf_path, storage_conf_path)
    else:
        raise Exception(f"Unsupported destination type: {destination_config['type']}")
    if 'concurrency' in profile:
        concurrency = profile['scheduler_concurrency']
    if 'quit_if_fail' in profile:
        quit_if_fail = profile['quit_if_fail']
    if 'thread_concurrency' in profile:
        thread_concurrency = profile['thread_concurrency']
    if args['test']:
        test(source, destination)
    elif args['meta']:
        meta(source, args['<meta_out_file_path>'])
    elif args['schema']:
        schema(source, destination, db_list, config_table_list, external_table_list, project_name, concurrency,
               quit_if_fail, thread_concurrency)
    elif args['data']:
        data(source, destination, db_list, config_table_list, external_table_list, project_name, concurrency,
             quit_if_fail, thread_concurrency, transform_partitions, dest_table_list)
    elif args['validate']:
        validate(source, destination, db_list, config_table_list, external_table_list, project_name, concurrency,
                 quit_if_fail, thread_concurrency)
    elif args['run']:
        run(source, destination, db_list, config_table_list, external_table_list, project_name, concurrency,
            quit_if_fail, thread_concurrency, transform_partitions)
    elif args['status']:
        status(destination, project_name)
    elif args['resume']:
        resume(source, destination)
    elif args['report']:
        report(destination, project_name)
    else:
        print("Unknown command")


if __name__ == '__main__':
    main()



import logging

import migration.util.migration_tasks_status as migration_tasks_status
from migration.base import ProfileConfigError
from migration.base.exceptions import SchedulerRuntimeError
from migration.base.status import TaskType
from migration.connector.destination.base import Destination
from migration.connector.source import Source
from migration.scheduler.scheduler import SchedulerConfig, Scheduler
from migration.scheduler.transformer import Transformer
from migration.scheduler.task.data_task import DataMigrationTask

DOT_SPLITTER = '.'
LEFT_BRACKET = '('

logger = logging.getLogger(__name__)


class DataTransformer(Transformer):
    def __init__(self, source: Source, destination: Destination, project_name: str, db_list=None,
                 config_table_list=None,
                 external_table_list=None,
                 scheduler_concurrency=1, quit_if_fail=False, thread_concurrency=1, transform_partitions=None, dest_table_list=None):
        super().__init__(source, destination, project_name, db_list, config_table_list, external_table_list,
                         scheduler_concurrency, quit_if_fail, thread_concurrency, dest_table_list)
        self.transform_partitions = transform_partitions

    def get_migration_tasks(self):
        transform_table_list = self.get_transform_table_list()
        if not transform_table_list:
            logger.warning("No table to transform, please check your profile.yml")
            raise ProfileConfigError("No table to transform, please check your profile.yml")
        data_migration_tables = []
        dest_table_list = self.get_dest_table_list()
        if dest_table_list:
            if len(dest_table_list) != len(transform_table_list):
                raise ProfileConfigError("dest_table_list length should be equal to transform_table_list length")

        if DOT_SPLITTER not in transform_table_list[0]:
            for db in transform_table_list:
                tables = self.source.get_table_names(db)
                for table in tables:
                    data_migration_tables.append(f"{db}.{table[0]}")
        else:
            for table in transform_table_list:
                if '*' in table:
                    db = table.split(DOT_SPLITTER)[0]
                    tables = self.source.get_table_names(db)
                    for result in tables:
                        data_migration_tables.append(f"{db}.{result[0]}")
                else:
                    data_migration_tables.append(table)
        logger.debug(f"data migration tasks: {data_migration_tables}")
        logger.info(f"data migration tasks count: {len(data_migration_tables)}")
        data_migration_tasks = []
        for index, table in enumerate(data_migration_tables):
            task = DataMigrationTask(name=table, task_type=TaskType.DATA_MIGRATION, destination=self.destination,
                                     project_id=self.project_id, source=self.source, dest_table=dest_table_list[index] if dest_table_list else None,
                                     transform_partitions=self.transform_partitions[
                                         table] if self.transform_partitions else None)
            migration_tasks_status.init_task_status(self.destination, task)
            data_migration_tasks.append(task)

        return data_migration_tasks

    def schedule_migration_tasks(self):
        migration_tasks = self.get_migration_tasks()
        if migration_tasks:
            if len(migration_tasks) < self.transform_concurrency:
                schedule_config = SchedulerConfig(quit_if_failed=self.quit_if_fail, max_workers=self.thread_concurrency)
                scheduler = Scheduler(schedule_config)
                self.schedules_tasks_map[scheduler] = migration_tasks
            else:
                for i in range(self.transform_concurrency):
                    schedule_config = SchedulerConfig(quit_if_failed=self.quit_if_fail, max_workers=self.thread_concurrency)
                    scheduler = Scheduler(schedule_config)
                    self.schedules_tasks_map[scheduler] = migration_tasks[i::self.transform_concurrency]



from enum import Enum


class Status(Enum):
    SUCCESS = 0,
    FAILED = 1,
    ERROR = 2,
    UNKNOWN = 3


class TaskStatus(Enum):
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"
    UNKNOWN = "UNKNOWN"
    INIT = "INIT"


class TaskType(Enum):
    SCHEMA_MIGRATION = "SCHEMA_MIGRATION"
    DATA_MIGRATION = "DATA_MIGRATION"
    DATA_VALIDATION = "DATA_VALIDATION"



import logging
import os
import subprocess
from time import sleep
from migration.connector.source.enum import Column

from migration.util import object_storage_util

from migration.connector.destination.base import Destination
from clickzetta.dbapi.connection import Connection as ClickZettaConnection
from clickzetta.client import Client
from migration.base.exceptions import DestinationExecutionError, GrammarRestrictionsError
import migration.util.script_util as script_utils

logger = logging.getLogger(__name__)


class ClickZettaDestination(Destination):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('Clickzetta', config)
        self.connection = None
        self.instance_id = None
        self.meta_conf_path = meta_conf_path
        self.storage_conf_path = storage_conf_path

    def get_connection_params(self):
        assert self.config.get("service"), "service is required"
        assert self.config.get("workspace"), "workspace is required"
        assert self.config.get("instance"), "instance is required"
        assert self.config.get("vcluster"), "vcluster is required"
        assert self.config.get("username"), "username is required"
        assert self.config.get("password"), "password is required"
        assert self.config.get("instanceId"), "instanceId is required"
        self.instance_id = self.config.get("instanceId")
        return {
            "service": self.config.get("service"),
            "workspace": self.config.get("workspace"),
            "instance": self.config.get("instance"),
            "vcluster": self.config.get("vcluster"),
            "username": self.config.get("username"),
            "password": self.config.get("password"),
        }

    def connect(self):
        if self.connection is None:
            connection_params = self.get_connection_params()
            client = Client(
                service=connection_params["service"],
                workspace=connection_params["workspace"],
                instance=connection_params["instance"],
                vcluster=connection_params["vcluster"],
                username=connection_params["username"],
                password=connection_params["password"],
            )
            self.connection = ClickZettaConnection(client=client)

    def create_table(self, table_name, ddl: str):
        self.execute_sql(ddl)
        logger.info(f"Create table {table_name} successfully")

    def get_table_columns(self, database_name, table_name) -> list[Column]:
        result = self.execute_sql(f"desc {database_name}.{table_name}")
        table_columns = []
        for row in result:
            column_name = row[0]
            column_type = row[1].upper() if 'not null' not in row[1] else row[1].strip().split(' ')[0].upper()
            is_null = False if 'not null' in row[1] else True
            table_columns.append(Column(column_name, column_type, is_null, default_value=None))
        return table_columns

    def create_database(self, db_name):
        self.execute_sql(f"CREATE SCHEMA IF NOT EXISTS {db_name}")
        logger.info(f"Create database {db_name} successfully")

    def execute_sql(self, sql, bind_params=None):
        try:
            self.connect()
            cursor = self.connection.cursor()
            cursor.execute(sql, bind_params)
            return cursor.fetchall()
        except Exception as e:
            logger.error("migration Error running SQL: {}, error:{}", sql, e)
            raise DestinationExecutionError('migration Error running SQL: {}, error:{}'.format(sql, e))

    def close(self):
        self.connection.close()

    def load_external_data(self, file_path, schema_name, table_name):
        meta_cmd_path = script_utils.get_meta_cmd_path()
        object_storage = object_storage_util.get_object_storage_config(self.storage_conf_path)
        object_file_path = file_path
        if object_storage['type'] == "oss":
            object_file_path = f"oss://{object_storage['bucket']}/{file_path}"
        elif object_storage['type'] == "cos":
            object_file_path = f"cos://{object_storage['bucket']}/{file_path}"
        else:
            logger.error(f"Unsupported object storage type {object_storage['type']}")
            raise DestinationExecutionError(f"Unsupported object storage type {object_storage['type']}")

        column_info = self.execute_sql(f"DESCRIBE {schema_name}.{table_name}")
        temp_table_name = f"temp_{table_name}"
        create_temp_table_sql = f"CREATE TABLE IF NOT EXISTS {schema_name}.{temp_table_name} (\n"
        for index, column in enumerate(column_info):
            create_temp_table_sql += f"    {column[0]} {column[1]}"
            if index < len(column_info) - 1:
                create_temp_table_sql += ",\n"
            else:
                create_temp_table_sql += "\n"
        create_temp_table_sql += (") using text options('field.delim' = '\t', 'serialization.null.format'= '\\N', "
                                  "'collection.delim'=':', 'mapkey.delim'='='); ")
        self.execute_sql(f"DROP TABLE IF EXISTS {schema_name}.{temp_table_name}")
        self.execute_sql(create_temp_table_sql)

        load_cmd = f"{meta_cmd_path} append-table -i {self.instance_id} -n {self.config.get('workspace')}" \
                   f",{schema_name} -t {temp_table_name} -r \"{object_file_path}\" -c {self.meta_conf_path}"
        logger.info(
            f"Clickzetta begin to load external data {object_file_path}"
            f" to {schema_name}.{table_name} with command: {load_cmd}")
        try:
            result = subprocess.run(load_cmd, shell=True, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, encoding="utf-8", timeout=300)
            if result.returncode != 0:
                raise DestinationExecutionError(
                    f"migration Error loading external data: {load_cmd}, error:{result.stderr}")

        except BaseException as e:
            logger.error("migration Error loading external data: {}, error:{}", load_cmd, e)
            raise DestinationExecutionError('migration Error loading external data: {}, error:{}'.format(load_cmd, e))
        rewrite_table_sql = f"INSERT INTO {schema_name}.{table_name} SELECT * FROM {schema_name}.{temp_table_name};"
        try:
            self.execute_sql(rewrite_table_sql)
            self.execute_sql(f"DROP TABLE IF EXISTS {schema_name}.{temp_table_name}")
        except BaseException as e:
            logger.error("migration Error loading external data: {}, error:{}", rewrite_table_sql, e)
            raise DestinationExecutionError(
                'migration Error loading external data: {}, error:{}'.format(rewrite_table_sql, e))
        logger.info(f"Clickzetta Load external data {file_path} to {schema_name}.{table_name} successfully")

    def gen_destination_ddl(self, db_name, table_name, columns, primary_keys, cluster_info, partition_columns):
        if primary_keys and cluster_info:
            for primary_key, cluster_key in zip(primary_keys, cluster_info.cluster_keys):
                if primary_key != cluster_key:
                    logger.warning(f"Primary key {primary_key} is not in cluster key {cluster_info.cluster_keys}")
                    raise GrammarRestrictionsError(f"Primary key {primary_key} is not in cluster key "
                                                   f"{cluster_info.cluster_keys}")
        destination_ddl = 'CREATE TABLE IF NOT EXISTS {db}.{table} (\n'.format(db=db_name, table=table_name)
        for index, column in enumerate(columns):
            destination_ddl += '    {name} {type} {is_null} ' \
                .format(name=column.name,
                        type=column.type,
                        is_null='' if column.is_null else 'NOT NULL')
            # if column.default_value:
            #     destination_ddl += "   DEFAULT {default_value}".format(default_value=column.default_value)
            if index != len(columns) - 1:
                destination_ddl += ',\n'
            else:
                destination_ddl += '\n'

        if primary_keys:
            destination_ddl += '    PRIMARY KEY ({primary_keys}),\n'.format(primary_keys=','.join(primary_keys))
        destination_ddl += ') \n'
        if cluster_info:
            destination_ddl += 'CLUSTERED BY ({clustered_by})\n'.format(
                clustered_by=','.join(cluster_info.cluster_keys))
            destination_ddl += 'SORTED BY ({clustered_by})\n'.format(clustered_by=','.join(cluster_info.cluster_keys))
            if cluster_info.bucket_num:
                destination_ddl += 'INTO {buckets} BUCKETS\n'.format(buckets=cluster_info.bucket_num)
            else:
                destination_ddl += 'INTO 32 BUCKETS\n'

        if partition_columns:
            destination_ddl += 'PARTITIONED BY ({partitioned_by})\n'.format(partitioned_by=','.join(partition_columns))
        self.create_database(db_name)
        return destination_ddl



import logging
import psycopg2

import psycopg2.pool as psycopg2_pool

from migration.connector.destination.base import Destination
from migration.base.exceptions import DestinationExecutionError, GrammarRestrictionsError, SourceExecutionError

logger = logging.getLogger(__name__)

MAX_GET_STATUS_TIMES = 100


class PGDestination(Destination):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('PostgreSQL', config)
        self.connection = None
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_conf_path = storage_conf_path
        pool_config = {'host': self.connection_params['host'], 'port': self.connection_params['port'],
                       'user': self.connection_params['user'], 'password': self.connection_params['password']}
        self.pool = psycopg2_pool.SimpleConnectionPool(minconn=10, maxconn=30, **pool_config)

    def get_connection_params(self):
        assert self.config['host']
        assert self.config['user']
        assert self.config['password']
        assert self.config['port']

        return {
            'host': self.config['host'],
            'port': int(self.config['port']),
            'user': self.config['user'],
            'password': self.config['password']
        }

    def connect(self):
        if self.connection is None:
            self.connection = psycopg2.connect(**self.connection_params)
            logger.info(f"Connect to PG {self.connection_params['host']} successfully")

    def execute_sql(self, sql, bind_params=None):
        connection = self.pool.getconn()
        try:
            with connection.cursor() as cur:
                cur.execute(sql, bind_params)
                result = cur.fetchall()
                return result

        except SourceExecutionError as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"
        except Exception as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"
        finally:
            self.pool.putconn(connection)

    def close(self):
        if self.connection is not None:
            self.connection.close()
        self.pool = None

    def quote_character(self):
        return "\""



import logging
import re
import time
from datetime import datetime
import psycopg2
from psycopg2 import pool as psycopg2_pool

from migration.connector.source.enum import Column

from migration.connector.source.base import Source
from migration.base.exceptions import SourceExecutionError

logger = logging.getLogger(__name__)


class PGSource(Source):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('PostgreSQL', config)
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_config_path = storage_conf_path
        pool_config = {'host': self.connection_params['host'], 'port': self.connection_params['port'],
                       'user': self.connection_params['user'], 'password': self.connection_params['password'],
                       'database': self.connection_params['database']
                       }
        self.pool = psycopg2_pool.SimpleConnectionPool(minconn=10, maxconn=30, **pool_config)

    """
    PostgreSQL connection parameters is a dict including following keys:
    1. host: host of PG
    2. port: port for PG client
    3. database: database name
    4. user: user name
    5. passwd: password
    """

    def get_connection_params(self):
        assert self.config['host']
        assert self.config['user']
        assert self.config['password']
        assert self.config['port']

        return {
            'host': self.config['host'],
            'port': int(self.config['port']),
            'user': self.config['user'],
            'password': self.config['password'],
            'database': self.config['database'],
        }

    def connect(self):
        if self.connection is None:
            self.connection = psycopg2.connect(**self.connection_params)
            logger.info(f"Connect to PG {self.connection_params['host']} successfully")

    def get_table_columns(self, database_name, table_name) -> list[Column]:
        result = self.execute_sql(f"select column_name, data_type, is_nullable,column_default "
                                  f"from INFORMATION_SCHEMA.COLUMNS where table_name =  '{table_name}' and table_schema = '{database_name}'")
        table_columns = []
        for row in result:
            table_columns.append(Column(name=row[0], type=row[1].upper(),
                                        is_null=True if row[2].strip() == 'YES' else False,
                                        default_value=row[3]))
        return table_columns

    def execute_sql(self, sql, bind_params=None):
        connection = self.pool.getconn()
        results = []
        try:
            with connection.cursor() as cur:
                cur.execute(sql)
                if sql.strip().upper().startswith('CREATE') or sql.strip().upper().startswith('DROP'):
                    return []
                for row in cur:
                    results.append(row)
                return results

        except SourceExecutionError as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"

        except BaseException as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"

        except Exception as e:
            logger.error(f"PG connector execute sql {sql} failed, error: {e}")
            raise f"PG connector execute sql {sql} failed, error: {e}"
        finally:
            self.pool.putconn(connection)

    def type_mapping(self):
        return {
            'TIME': 'DATE',
            'CHARACTER': 'VARCHAR',
            'CHARACTER VARIATION': 'VARCHAR',
            'REAL': 'FLOAT',
            'DOUBLE PRECISION': 'DOUBLE',
            'INTEGER': 'INT',
            'TIMESTAMP WITHOUT TIME ZONE': 'TIMESTAMP',
        }

    def close(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None
        if self.pool is not None:
            self.pool = None
        logger.info(f"Close connection to PG {self.connection_params['host']} successfully")

    def int_type_string(self):
        return 'INT'

    def quote_character(self):
        return "\""

    def get_table_pk_columns(self, database_name, table_name):
        result = self.execute_sql(f"select column_name from INFORMATION_SCHEMA.COLUMNS where table_name =  '{table_name}' and table_schema = '{database_name}' and column_key = 'PRI'")
        pk_columns = [row[0] for row in result]
        return tuple(pk_columns)


import logging
import os
import random
from datetime import datetime
from time import sleep

import pymysql
import pymysqlpool as pymysql_pool

from migration.connector.destination.base import Destination
from clickzetta.dbapi.connection import Connection as ClickZettaConnection
from clickzetta.client import Client
from migration.base.exceptions import DestinationExecutionError, GrammarRestrictionsError
from migration.base.status import Status
import migration.util.object_storage_util as object_storage_util

logger = logging.getLogger(__name__)

MAX_GET_STATUS_TIMES = 100


class MysqlDestination(Destination):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('Mysql', config)
        self.connection = None
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_conf_path = storage_conf_path
        pool_config = {'host': self.connection_params['host'], 'port': self.connection_params['port'],
                       'user': self.connection_params['user'], 'password': self.connection_params['password']}
        self.pool = pymysql_pool.ConnectionPool(size=10, maxsize=50, pre_create_num=5, name='mysql_pool', **pool_config)

    def get_connection_params(self):
        assert self.config['host']
        assert self.config['user']
        assert self.config['password']
        assert self.config['port']

        return {
            'host': self.config['host'],
            'port': int(self.config['port']),
            'user': self.config['user'],
            'password': self.config['password']
        }

    def connect(self):
        if self.connection is None:
            self.connection = pymysql.connect(**self.connection_params)
            logger.info(f"Connect to Mysql {self.connection_params['host']} successfully")

    def create_table(self, table_name, ddl: str):
        self.execute_sql(ddl)
        logger.info(f"Create table {table_name} successfully")

    def create_database(self, db_name):
        self.execute_sql(f"CREATE DATABASE IF NOT EXISTS {db_name}")
        logger.info(f"Create database {db_name} successfully")

    def execute_sql(self, sql, bind_params=None):
        connection = self.pool.get_connection()
        try:
            with connection.cursor() as cur:
                cur.execute(sql, bind_params)
                result = cur.fetchall()
                return result
        except Exception as e:
            logger.error("migration Error running SQL: {}, error:{}", sql, e)
            raise DestinationExecutionError('migration Error running SQL: {}, error:{}'.format(sql, e))

        finally:
            connection.close()

    def close(self):
        if self.connection is not None:
            self.connection.close()
        self.pool = None


import logging
import os

logger = logging.getLogger(__name__)


def get_meta_cmd_path():
    if os.name == 'nt':
        meta_cmd_path = os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
                                     'scripts/linux/meta_cmd.exe')
        logger.info(f'meta_cmd_path: {meta_cmd_path}')
        return meta_cmd_path
    elif os.name == 'posix':
        meta_cmd_path = os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
                                     'scripts/linux/meta_cmd')
        logger.info(f'meta_cmd_path: {meta_cmd_path}')
        return meta_cmd_path
    elif os.name == 'mac':
        meta_cmd_path = os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
                                     'scripts/linux/meta_cmd')
        logger.info(f'meta_cmd_path: {meta_cmd_path}')
        return meta_cmd_path
    else:
        raise Exception(f'Unsupported OS:{os.name}')


def get_files_path(instance_id, workspace, schema, table, meta_conf_path):
    get_files_cmd = f'{get_meta_cmd_path()} list-data-files -i {instance_id} -n {workspace},{schema} -t {table} -c {meta_conf_path}'
    logger.info(f'get_files_cmd: {get_files_cmd}')
    try:
        result = os.popen(get_files_cmd).readlines()
    except Exception as e:
        logger.error(f'get_files_cmd error: {e}')
        raise Exception(f'get_files_cmd error: {e}')
    files = set()
    for line in result:
        if line.startswith('file_path'):
            last_item_length = len(line.replace('\"', '').split('/')[-1])
            file_path = line.replace('\"', '')[:-last_item_length].split('/', 3)[-1]
            logger.info(f'Add Object storage file_path: {file_path}')
            files.add(file_path)
    return files



import logging
from datetime import datetime

from migration.base.exceptions import DestinationExecutionError
from migration.scheduler.task.base_task import Task, TaskType, TaskStatus
from migration.connector.destination.base import Destination
import migration.util.migration_tasks_status as migration_tasks_status

logger = logging.getLogger(__name__)


class SchemaMigrationTask(Task):

    def __init__(self, name: str, task_type: TaskType, ddl: str, destination: Destination, *args, **kwargs):
        super().__init__(name, task_type, *args, **kwargs)
        self.ddl = ddl
        self.destination = destination

    def run(self) -> Task:
        try:
            self.status = TaskStatus.RUNNING
            self.destination.execute_sql(self.ddl)
            self.status = TaskStatus.COMPLETED
        except BaseException as e:
            self.status = TaskStatus.FAILED
            self.end_time = datetime.now()
            migration_tasks_status.update_task_status(self.destination, self)
            logger.error(f"SchemaMigrationTask {self.name} failed to run, error: {e}, retry times: {self.retry_times}")
            return self
        self.end_time = datetime.now()
        migration_tasks_status.update_task_status(self.destination, self)
        logger.info(f"SchemaMigrationTask {self.name} finished running, status: {self.status.value}")

        return self



import logging
import time
from concurrent.futures import ThreadPoolExecutor, CancelledError
import threading
from queue import Queue, Empty
import schedule
from migration.base.exceptions import TaskRuntimeError

logger = logging.getLogger(__name__)


class SchedulerConfig:
    def __init__(self, max_workers=10, queue_size=10, task_timeout=60, task_retry=3, quit_if_failed=False):
        self.max_workers = max_workers
        self.queue_size = queue_size
        self.task_timeout = task_timeout
        self.task_retry = task_retry
        self.quit_if_failed = quit_if_failed


class Scheduler:
    def __init__(self, config: SchedulerConfig):
        self.config = config
        self.executor = ThreadPoolExecutor(max_workers=config.max_workers)
        self.future_results = Queue(maxsize=config.queue_size)
        self.deamon_thread = threading.Thread(daemon=True, target=self.run_scheduler_jobs)
        self.succeed_tasks = []
        self.failed_tasks = []
        self.running_tasks = []
        self.finish_called = False
        self.tasks_num = 0
        logger.info(f"Scheduler is initialized with config: {config.__dict__}")

    def add_task(self, task):
        self.future_results.put(self.executor.submit(task.run))
        logger.info(f"Task:{task.__str__} is added to scheduler")

    def run(self):
        try:
            schedule.every(1).seconds.do(self.check_task_result)
            schedule.every(1).seconds.do(self.check_running_task_status)
            self.deamon_thread.start()
        except Exception as e:
            logger.error(f"Scheduler failed to process all tasks, error: {e}")

    def run_scheduler_jobs(self):
        try:
            while True:
                if self.finish_called:
                    break
                schedule.run_pending()
                time.sleep(1)
        except Exception as e:
            logger.error(f"Scheduler failed to process all tasks, error: {e}")

    def check_task_result(self):
        if self.future_results.empty():
            return
        try:
            feature = self.future_results.get_nowait()
            if feature is not None:
                if feature.cancelled():
                    logger.error("Task is cancelled with unknown reason")
                    raise CancelledError("Task is cancelled with unknown reason")

                elif feature.done():
                    self.process_done_task_feature(feature)
                elif feature.running():
                    print("add running task")
                    self.running_tasks.append(feature)
                elif feature._state == "PENDING":
                    print("add pending task")
                    self.running_tasks.append(feature)

        except CancelledError as e:
            raise e
        except Exception as e:
            raise e

    def shutdown(self):
        self.executor.shutdown(wait=True, cancel_futures=True)
        schedule.clear()
        self.deamon_thread.join()
        for succeed_task in self.succeed_tasks:
            logger.info(f"Task:{succeed_task.__str__} is succeed")
        for failed_task in self.failed_tasks:
            logger.error(f"Task:{failed_task.__str__} is failed")

    def process_done_task_feature(self, feature):
        try:
            result = feature.result()
            if result.is_success():
                self.succeed_tasks.append(result)
            elif result.is_failed():
                if result.retry_times < self.config.task_retry:
                    logger.info(
                        f"Task:{result.__str__} is failed, begin to retry the {result.retry_times + 1} time, "
                        f"max_retry is {self.config.task_retry}")
                    result.retry_times += 1
                    self.future_results.put(self.executor.submit(result.run))
                else:
                    logger.error(f"Task:{result.__str__} is failed, max_retry is {self.config.task_retry}")
                    self.failed_tasks.append(result)
        except TaskRuntimeError as e:
            raise e

    def check_running_task_status(self):
        for task in self.running_tasks:
            if task.done():
                self.process_done_task_feature(task)
                self.running_tasks.remove(task)
                print(f"remove running task {task} from running_tasks")

    def finish(self):
        while True:
            logger.info(
                f"succeed_task_count:{self.succeed_tasks.__len__()}, "
                f"failed_task_count:{self.failed_tasks.__len__()}, total_task_count:{self.tasks_num}")
            if self.succeed_tasks.__len__() + self.failed_tasks.__len__() == self.tasks_num:
                self.finish_called = True
                self.shutdown()
                break
            else:
                logger.info(
                    f"Waiting for {len(self.running_tasks)} tasks to finish")
                time.sleep(5)
